from typing import *
from typing import BinaryIO
from enum import Enum
import os
from datalogue.utils import Json
from datalogue.errors import DtlError
import logging
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from requests import Response
from datalogue.models.organization import _users_from_payload
from pathlib import Path

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

logger = logging.getLogger("_HttpClient")


class HttpMethod(Enum):
    POST = 'POST'
    GET = 'GET'
    PUT = 'PUT'
    DELETE = 'DELETE'


class MultiPartFile:
    def __init__(self, name: str, file: BinaryIO):
        self.name = name
        self.file = file


class MultiPartData:
    def __init__(self, name: str, data: str):
        self.name = name
        self.data = data


def _has_ssl_certificate():
    """
    Check if SSL certificate is installed, prints to logger if not verified, and returns the certificate.
    :return: True if contains SSL certificate, False otherwise
    """
    verify_cert_env = os.environ.get('DTL_SSL_VERIFY_CERT')
    verify_certificate = False if (verify_cert_env == 'false') else True
    if not verify_certificate:
        logger.info('SSL Certification verification is disabled.')
    return verify_certificate


def _check_get_ca_certificate_path(ca_certificate: Path):
    """
    If not None, check if ca_certificate is indeed a path
    :param ca_certificate:
    :return: ca_certificate path, DtlError if the ca_certificate path is not found, True if contains SSL certificate or
    False if it doesn't
    """
    if ca_certificate is None:
        verify_certificate = _has_ssl_certificate()
    else:
        if not os.path.exists(ca_certificate):
            return DtlError("'%s' was not found. Please, check the path to the certificate." % ca_certificate)
        verify_certificate = ca_certificate
    return verify_certificate


class _HttpClient:
    """
    Class that Abstracts aways requests and makes it use type safe.
    It also prepares all the requests to be used authenticated to the API
    """

    def __init__(self, uri: str):
        self.uri = uri
        self.http_session = requests.Session()

    def get_csrf(self):
        public_url = self.uri
        if os.environ.get('DTL_PUBLIC_URL') is not None:
            public_url = os.environ.get('DTL_PUBLIC_URL')

        self.http_session.headers.update({
            'Origin': public_url.strip("/api")
        })

        url = self.uri + '/health'
        res = self.http_session.get(url)

        if "csrf-token" in res.headers:
            self.http_session.cookies.set('csrf-token', res.headers["csrf-token"])
            self.http_session.headers.update({
                'X-Csrf-Token': res.headers["csrf-token"]
            })

    def login(self, username: str, password: str, ca_certificate: Optional[str] = None) -> Union[DtlError, None]:
        """
        Sets the cookie for the client to use

        :param username: Username of user to login
        :param password: Password of user to login
        :param ca_certificate: (Optional) path to the CA Certificate
        :return: Nothing
        """
        self.get_csrf()

        verify_certificate = _check_get_ca_certificate_path(ca_certificate)
        if isinstance(verify_certificate, DtlError):
            return verify_certificate

        url = self.uri + '/signin'
        try:
            res = self.http_session.post(
                url,
                json={
                    'email': username,
                    'password': password
                },
                allow_redirects=False,
                verify=verify_certificate
            )
        except requests.ConnectionError:
            return DtlError("There was a connection error trying to reach: %s" % self.uri)
        except requests.HTTPError:
            return DtlError("There was an HTTP error trying to reach: %s" % self.uri)
        except requests.URLRequired:
            return DtlError("%s is not a valid endpoint." % self.uri)
        except requests.Timeout:
            return DtlError("The request timed out")

        if res.status_code != 201 and res.status_code != 302 and res.status_code != 200:
            return DtlError('Authentication failed - %s : %s' % (str(res.status_code), res.text))
        elif res.status_code == 503 or res.status_code == 504:
            return DtlError("The service is not available. StatusCode: %s, Body: %s" % (str(res.status_code), res.text))
        else:
            return None

    def signup(self, first_name, last_name, email, password, accept_terms, ca_certificate: Optional[str] = None):
        """
        Allows
        :param first_name: First name of user to be registered
        :param last_name: Last name of user to be registered
        :param email: Email of user to be registered
        :param password: Password of user to be registered
        :param accept_terms: Whether user accepts the term of service here : https://assets.website-files.com/5c980fa2de94e8f79f047ded/5cd0e4d60880b51cb361ae7a_Terms%20of%20Service.pdf
        :param ca_certificate: (Optional) path to the CA Certificate
        :return: A User object
        """
        print("By calling this action, you've agreed to the terms of service as set forth here : "
                   "https://assets.website-files.com/5c980fa2de94e8f79f047ded/5cd0e4d60880b51cb361ae7a_Terms%20of%20Service.pdf")

        verify_certificate = _check_get_ca_certificate_path(ca_certificate)
        if isinstance(verify_certificate, DtlError):
            return verify_certificate

        if not accept_terms:
            return DtlError("Terms & Conditions for registration are not accepted")
        url = self.uri + '/signup'
        self.get_csrf()

        try:
            res = self.http_session.post(url, json={
                    'email': email,
                    'firstName': first_name,
                    'lastName': last_name,
                    'password': password
                }, verify=verify_certificate)
        except requests.ConnectionError:
            return DtlError("There was a connection error trying to reach: %s" % self.uri)
        except requests.HTTPError:
            return DtlError("There was an HTTP error trying to reach: %s" % self.uri)
        except requests.URLRequired:
            return DtlError("%s is not a valid endpoint." % self.uri)
        except requests.Timeout:
            return DtlError("The request timed out")

        if res.status_code == 200:
            if res.headers.get("content-type") == "application/json":
                return _users_from_payload(res.json())
            else:
                return DtlError("Cannot parse response, unknown type")
        elif res.status_code == 503:
            return DtlError("The service is not available")
        else:
            return DtlError(res.text)

    @staticmethod
    def _handle_status_code(request_uri: str, res: Response) -> Union[DtlError, None]:
        """
        Return a proper DtlError depending on the status code of the Response

        :param request_uri: uri of th request that was made
        :param res: Response received to the request
        :return: an Error or None if the status doesn't warrant returning an error
        """
        if res.status_code == 400:
            return DtlError("Invalid Request to %s returned 400, body res: %s" % (request_uri, res.text))

        if res.status_code == 401:
            return DtlError("It looks like you are not logged in, please login")

        if res.status_code == 403:
            return DtlError(f"Authorization error. Make sure you're authorized to do this operation. Msg: {res.text}")

        if res.status_code == 404:
            return DtlError("Route not found %s" % request_uri)

        if res.status_code == 413:
            return DtlError("Body is too big, please limit uploads to 1 Go")

        if res.status_code == 500:
            if not res.text:
                return DtlError("Internal Server error")
            else:
                return DtlError(res.text)

        if res.status_code == 503:
            return DtlError("The service is not available")

        return None

    def post_multipart_data(self, path: str, files: List[MultiPartFile], data: List[MultiPartData]) -> Union[DtlError, Response]:
        headers = { 'content-type': "multipart/form-data" }
        
        files_payload = {}
        for f in files:
            files_payload[f.name] = f.file

        data_payload = {}
        for d in data:
            data_payload[d.name] = d.data
        
        request_uri = self.uri + path
        req = requests.Request(
            HttpMethod.POST.value, 
            request_uri, 
            files=files_payload, 
            data=data_payload)
        prepped = self.http_session.prepare_request(req)
        res = self.http_session.send(prepped)

        e = self. _handle_status_code(request_uri, res)
        if isinstance(e, DtlError):
            return e
        
        content_type_header = res.headers.get("content-type")
        if len(res.text) > 0 and content_type_header == "application/json":
            return res.json()
        else:
            return res.text

    def post_binary(self, path: str, file: BinaryIO, params: Optional[dict] = None) -> Union[DtlError, Response]:
        """
        Makes a post request with the content being a binary file that will be uploaded chunk by chunk

        :param path: path uri to the datastore
        :param file: iterator that is going to generate the binary data
        :param params: parameters to be used in the request
        :return:
        """
        request_uri = self.uri + path
        req = requests.Request(HttpMethod.POST.value, request_uri, data=file, params=params)
        prepped = self.http_session.prepare_request(req)

        prepped.headers['Content-Type'] = 'application/octet-stream'

        res = self.http_session.send(prepped)

        e = _HttpClient._handle_status_code(request_uri, res)
        if isinstance(e, DtlError):
            return e

        return res

    def make_authed_request(self, path: str, method: HttpMethod, body: Optional[dict] = None,
                            params: Optional[dict] = None, stream=False, verify=True) -> Union[DtlError, Json, Response]:
        res = self.execute_authed_request(path, method, body, params, stream, verify)
        request_uri = self.uri + path
        e = self._handle_status_code(request_uri, res)
        if isinstance(e, DtlError):
            return e

        content_type_header = res.headers.get("content-type")
        if stream:
            return res
        elif len(res.text) > 0 and content_type_header == "application/json":
            return res.json()
        else:
            return res.text

    def execute_authed_request(self, path: str, method: HttpMethod, body: Optional[dict] = None,
                               params: Optional[dict] = None, stream=False, verify=True) -> Union[DtlError, Response]:
        """
        Makes a request to the API give the url

        :param path: path to access the datastore below the host
        :param method: parameter to indicate HTTP Method to be used in the request
        :param body: body to be sent along with the request
        :param params: query parameters, to be fed with a dictionary
        :param stream: boolean to activate streaming the content of the response, this will delay the retrieval of
        the body until .content is called.
        :return: Request response in raw format
        """
        self.get_csrf()
        request_uri = self.uri + path
        req = requests.Request(method.value, request_uri, json=body, params=params)
        prepped = self.http_session.prepare_request(req)
        res = self.http_session.send(prepped, stream=stream, verify=verify)
        return res

    def put_binary(self, path: str, file: BinaryIO, params: Optional[dict] = None) -> Union[DtlError, Response]:
        """
        Makes a put request with the content being a binary file that will be uploaded chunk by chunk

        :param path: path uri to the datastore
        :param file: iterator that is going to generate the binary data
        :param params: parameters to be used in the request
        :return:
        """
        self.get_csrf()
        request_uri = self.uri + path
        req = requests.Request(HttpMethod.PUT.value, request_uri, data=file, params=params)
        prepped = self.http_session.prepare_request(req)

        prepped.headers['Content-Type'] = 'application/octet-stream'

        res = self.http_session.send(prepped)

        e = _HttpClient._handle_status_code(request_uri, res)
        if isinstance(e, DtlError):
            return e

        return res



import sys
from typing import Dict, Optional, Union, List
from uuid import UUID

import pandas as pd

from datalogue.errors import DtlError
from datalogue.models.graph import Graph, Node, NodeDef, Path
from datalogue.utils import _parse_string_list, _parse_list
from datalogue.utils import _parse_datetime, _parse_uuid, _parse_list
from datalogue.utils import map_option
from datetime import datetime
from datalogue.errors import _enum_parse_error
from datalogue.utils import SerializableStringEnum

class ClassificationType(SerializableStringEnum):
    Automatic = "Automatic"
    Manual = "Manual"

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("stream state", s))

    @staticmethod
    def classification_type_from_str(string: str) -> Union[DtlError, 'ClassificationType']:
        return SerializableStringEnum.from_str(ClassificationType)(string)

class Classification:
    """
    Represents the mapping and application of a class to a field. To edit classifications, please use the add, remove, and override commands of the dtl.datastore client. 
    """
    def __init__(self, class_id: UUID, classification_type: ClassificationType, class_name: str, class_path: List[str], score: float, created_at: datetime, created_by: Optional[UUID]):
        self.class_id = class_id
        self.classification_type = classification_type
        self.class_name = class_name
        self.class_path = class_path
        self.score = score
        self.created_at = created_at
        self.created_by = created_by

    def __repr__(self):
        return f"({self.class_name}, {self.score})"
    
    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, 'Classification']:
        class_id = map_option(payload.get("classId"), _parse_uuid) 
        if class_id is None:
            return DtlError("Classification has to have a 'class_id' key")

        if class_id is DtlError:
            return class_id

    
        classification_type = map_option(payload.get("classificationType"), ClassificationType.classification_type_from_str)
        if classification_type is None:
            return DtlError("Classification has to have a 'classification_type' key")

        if classification_type is DtlError:
            return classification_type

        score = payload.get("score")
        if score is None:
            return DtlError("SchemaNode has to have a 'score' key")

        created_at = map_option(payload.get("createdAt"), _parse_datetime)
        if isinstance(created_at, DtlError):
            return created_at

        created_by = map_option(payload.get("createdBy"), _parse_uuid)
        if isinstance(created_by, DtlError):
            return created_by

        return Classification(class_id, classification_type, None, None, score, created_at, created_by)

class SchemaNode(Node):
    """
    Describes schema node
    """

    def __init__(self, id: UUID, label: str, dataType: str, classifications: List[Classification]):
        """
        Builds SchemaNode object

        :param id: schema node id
        :param label: column name
        :param dataType: type of the data
        :param ontologyNodeId: user-classified ontology id for this node
        :param ontologyNodePath: user-classified ontology path for this node
        """

        self.id = id
        self.label = label
        self.dataType = dataType
        self.classifications = classifications
        Node.__init__(self, id, label)

    def __repr__(self):
        classificationStr = ", ".join(list(map(lambda c: str(c), self.classifications)))
        return ('SchemaNode('
                    f'id: {self.id}, '
                    f'label: {self.label}, '
                    f'dataType: {self.dataType}, '
                    f'\n\tclassifications: [{classificationStr}]\n)')

    @staticmethod
    def _as_payload(self) -> Union[DtlError, dict]:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """

        return {
            "id": self.id,
            "label": self.label,
            "dataType": self.dataType
        }

    def _schema_node_from_payload(json: dict) -> Union[DtlError, 'SchemaNode']:
        node_id = json.get("id")
        if node_id is None:
            return DtlError("SchemaNode has to have a 'id' key")

        label = json.get("label")
        if label is None:
            return DtlError("SchemaNode has to have a 'label' key")

        dataType = json.get("dataType")
        classifications = _parse_list(Classification._from_payload)(json.get("classifications"))
        if isinstance(classifications, DtlError):
            return classifications

        return SchemaNode(node_id, label, dataType, classifications)

class GraphSampleNode(Node):
    """
    Describes schema node
    """

    def __init__(self, id: UUID, label: str, value: str):
        """
        Builds SchemaNode object

        :param id: schema node id
        :param label: column name
        :param value: value of the data
        """

        self.id = id
        self.label = label
        self.value = value
        Node.__init__(self, id, label)

    def _sample_node_from_payload(json: dict) -> Union[DtlError, 'GraphSampleNode']:
        node_id = json.get("id")
        if node_id is None:
            return DtlError("SampleNode has to have a 'id' key")

        label = json.get("label")
        if label is None:
            return DtlError("SampleNode has to have a 'label' key")

        value = json.get("value")
        if "value" not in json:
            return DtlError("SampleNode has to have a 'value' key")

        return GraphSampleNode(str(node_id), label, value)



class AbstractDataGraph(Graph):
    node_decoder = GraphSampleNode._sample_node_from_payload


class AbstractDataSchema(Graph):
    node_decoder = SchemaNode._schema_node_from_payload



class SchemaOntologyRow:
    """
    Represents printable representation of schema, its associated ontology & samples data.
    """

    def __init__(self, schema_node_def: NodeDef, samples: List[AbstractDataGraph]):
        self.schema_node_def = schema_node_def
        self.samples = samples

    def to_array(self, is_full_path: bool):
        def get_sample_val(index: int, full_path: Path):
            if len(self.samples) > index:
                sample = self.samples[index]
                return sample.path_to_node[full_path].node.value
            else:
                return None

        schema_node = self.schema_node_def.node
        full_path = self.schema_node_def.full_path
        sample_1_val = get_sample_val(0, full_path)
        sample_2_val = get_sample_val(1, full_path)
        sample_3_val = get_sample_val(2, full_path)
        ontology_node_id = schema_node.ontologyNodeId
        if ontology_node_id is None:
            ontology_node_id = "\"\""
        if not is_full_path:
            if schema_node.ontologyNodePath is not None and len(schema_node.ontologyNodePath) > 0:
                ontology_node_label = schema_node.ontologyNodePath[-1]
            else:
                ontology_node_label = "\"\""
            return [
                schema_node.id, 
                schema_node.label, 
                schema_node.dataType,
                ontology_node_label, 
                ontology_node_id, 
                sample_1_val, 
                sample_2_val,
                sample_3_val
            ]
        else:
            if schema_node.ontologyNodePath is not None:
                ontology_node_label = schema_node.ontologyNodePath
            else:
                ontology_node_label = ["\"\"", "\"\""]
            return [
                schema_node.id, 
                self.schema_node_def.full_path, 
                schema_node.dataType,
                ontology_node_label, 
                ontology_node_id, 
                sample_1_val, 
                sample_2_val,
                sample_3_val
            ]


class Schema:
    """
    Represents a schema.
    """

    def __init__(self, data_schema: AbstractDataSchema, samples: Optional[List[AbstractDataGraph]]):
        """
        Represents a schema structure

        :param data_schema: Represents data schema
        :param samples: A collection of samples data with the same structure as data_schema
        """

        self.data_schema = data_schema
        self.samples = samples

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Schema']:
        schema = json.get("schema")
        samples_json = json.get("samples")
        if schema is None:
            return DtlError("Schema has to have a 'schema' key")
        if samples_json is None:
            return DtlError("Schema has to have a 'samples' key")
        ads = AbstractDataSchema._from_payload(schema, AbstractDataSchema.node_decoder)

        if samples_json is None:
            adg = None
        else:
            adg = []
            for obj in samples_json:
                parsed_obj = AbstractDataGraph._from_payload(obj, AbstractDataGraph.node_decoder)

                if isinstance(parsed_obj, DtlError):
                    return parsed_obj
                else:
                    adg.append(parsed_obj)
        return Schema(ads, adg)

    def print(self, full_path: bool):
        pd.set_option('max_colwidth', 29)
        nodes_def = self.data_schema.get_nodes()
        res = []
        if len(nodes_def) > 1000:
            return DtlError("Extremely large schema, please contact Datalogue support to enable larger schemas")
        for node_def in nodes_def:
            res.append(SchemaOntologyRow(node_def, self.samples).to_array(full_path))
        df = pd.DataFrame(pd.np.array(res),
                     columns=['schema_node_id', 'schema_node_name', 'schema_node_type', 'ontology_node_name',
                              'ontology_node_id', 'sample1', 'sample2', 'sample3'])
        sys.stdout.write(df.to_string(index=False))



from typing import List, Union, Optional
from datalogue.errors import DtlError, _property_not_found
from datalogue.models.transformations.commons import Transformation
from datalogue.models.classifier import ClassificationMethod, _classification_method_from_payload
from datalogue.utils import _parse_list
from uuid import UUID


# Kept because classify_fields is still depending on it
class Classifier(object):

    def __init__(self, classification_methods: List[ClassificationMethod], default_class_id: Optional[Union[UUID, str]] = None):
        """
        Build a local Classifier

        :param classification_methods: ordered list of Classification Methods
        :param default_class_id: allows the user to specify the class to be defaulted to, if all classification methods
            fail. Default value is `Unknown`. (currently part of replace_class, see below)
        """
        self.classification_methods = classification_methods
        self.default_class_id = default_class_id

    def __repr__(self):
        return ('Classifier('
               f'default_class: {self.default_class_id} '
               f'classification_methods: \n{self.classification_methods})')

    def __eq__(self, other: 'Classifier'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> dict:
        base = {
            "classificationMethods":  list(map(lambda m: m._as_payload(), self.classification_methods)),
            "defaultClass": self.default_class_id
        }
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Classifier']:
        classification_methods = json.get("classificationMethods")
        if classification_methods is None:
            return _property_not_found("classificationMethods", json)
        else:
            classification_methods = _parse_list(_classification_method_from_payload)(classification_methods)
            if isinstance(classification_methods, DtlError):
                return classification_methods

        default_class = json.get("default_class")

        return Classifier(classification_methods, default_class)


class Classify(Transformation):
    """
    Apply classes to all data streaming through a pipeline. Classification is based on the Classifier supplied.
    """
    type_str = "Classify"

    def __init__(self, classifier_id: Union[str, UUID], fields_to_target: Optional[List[List[str]]] = None,
                 add_class_fields=False, add_score_fields=False):
        """
        Unless specified with parameters below, this transformation does not alter the data schema; classes are
        represented in the backend only.

        :param classifier_id: id of the classifier to use for classification
        :param fields_to_target: specifies which fields to target for classification; remainder are not classified by
         this transformation
        :param add_class_fields: if True, adds a field containing class name for every classified field
        :param add_score_fields: if True, adds a field containing class score for every classified field
        """

        Transformation.__init__(self, Classify.type_str)
        self.classifier_id = classifier_id
        self.fields_to_target = fields_to_target
        self.add_class_fields = add_class_fields
        self.add_score_fields = add_score_fields

    def __eq__(self, other: 'Classify'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return('Classify('
               f'classifier_id: {self.classifier_id}, '
               f'fields_to_target: {self.fields_to_target}, '
               f'add_class_fields: {self.add_class_fields}, '
               f'add_score_fields: {self.add_score_fields})')

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["classifier"] = str(self.classifier_id)
        base["fieldsToTarget"] = self.fields_to_target
        base["addClassFields"] = self.add_class_fields
        base["addScoreFields"] = self.add_score_fields
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Classify']:
        classifier_id = json.get("classifier")
        if classifier_id is None:
            return _property_not_found("classifier", json)

        fields_to_target = json.get("fieldsToTarget")

        add_class_fields = json.get("addClassFields")
        if add_class_fields is None:
            add_class_fields = False

        add_score_fields = json.get("addScoreFields")
        if add_score_fields is None:
            add_score_fields = False

        return Classify(classifier_id, fields_to_target, add_class_fields, add_score_fields)
